'use strict'

const { program } = require('commander');
const fs = require('fs')
const diffParse = require('parse-diff')
const readCoverageFromXml = require('./cobertura-parse').readCoverageFromXml
const generateHtmlReport = require('./report').generateHtmlReport

const app = {
  run() {
    program
      .version('1.0.9')
      .description('Compare cobertura.xml generated by jest with a diff patch file')
      .option('-c, --cobertura <cobertura>', 'The file generated by the (feature) branch you want to check', 'coverage/cobertura.xml')
      .option('-d, --diff <diff>', 'The diff patch file ', 'diff.patch')
      .option('-o, --output <output>', 'The output file to contain the results', 'cobertura-coverage.html')
      .option('-p, --project <project>', 'The project path form the workspace root', '')
      .option('-t, --threshold <threshold>','The minimal threshold code should have', '0.70')
      .option('-v, --variance <variance>','Allow for variance, e.g. allow 5 to allow a 5% decrease', '0');
    program.parse(process.argv);
    const options = program.opts();

    console.info(`##Cobertura Diff Coverage ## Version: ${options['version']}`);

    this.validateOptions(options);

    // this example reads the file synchronously
    // you can read it asynchronously also
    let xmlString = fs.readFileSync(options.cobertura, "utf8");
    let diffstring = fs.readFileSync(options.diff, "utf8");

    var diffFile = diffParse(diffstring);

    let exitCode = 0;
    const minCov = parseFloat(options.threshold*100);

    const uncoveredFiles = new Object();

    if (options.variance!='0') {
      console.log(`Allowed variance is ${options.variance}`);
    }
    //Parse the cobertura.xml file content to loop through the relevant files.
    readCoverageFromXml(xmlString, {})
    .then(function (result) {
      try {
        var linesToCover = 0
        var coveredLines = 0
        var branchesToCover = 0
        var coveredBranches = 0
        result.files.forEach(file => {
          let fileLines = file.lines
          var fileLinesToCover = 0
          var fileCoveredLines = 0
          var fileBranchesToCover = 0
          var fileCoveredBranches = 0
          const fileUncoveredLines = new Array();
          const fileUncoveredBranches = new Array();

          //Locate the file chunks into the patch file
          let patchObject = diffFile.find(patch => patch.to === options.project+file.file)
          if(patchObject )
          {
            console.log(patchObject)
            //Flat the changes form the chunks and filter the additions
            let additions = patchObject.chunks.flatMap(change => change.changes).filter(change => change.add)
            additions.forEach(addition => {
              let diffLine = fileLines.find(line => parseInt(line.number) === addition.ln)
              if(diffLine) {
                if(diffLine.branch == 'true' && addition.content.indexOf('async') == -1)// workaround for issue with jest and async https://github.com/facebook/jest/issues/7829
                {
                  branchesToCover += diffLine.branchesToCover
                  coveredBranches += diffLine.branchesCovered
                  fileBranchesToCover += diffLine.branchesToCover
                  fileCoveredBranches += diffLine.branchesCovered
                  if(diffLine.branchesToCover > diffLine.branchesCovered) {
                    fileUncoveredBranches.push({line: diffLine.number, content: addition.content, missing: `(${diffLine.branchesCovered}/${diffLine.branchesToCover})`})
                  }
                }
                else {
                  linesToCover++
                  fileLinesToCover ++
                  coveredLines = coveredLines + (parseInt(diffLine.hits) > 0 ? 1 : 0)
                  fileCoveredLines = fileCoveredLines + (parseInt(diffLine.hits) > 0 ? 1 : 0)
                  if(!parseInt(diffLine.hits) > 0 ) {
                    fileUncoveredLines.push({line: diffLine.number, content: addition.content})
                  }
                }
              }
            })
            if(fileLinesToCover+fileBranchesToCover === 0) {
              console.log(`File ${file.file} has no lines to cover`)
            } else {
              console.log(`${file.file} coverage % is: ${(fileCoveredLines+fileCoveredBranches) / (fileLinesToCover + fileBranchesToCover) * 100}` )
              const overallCov = parseFloat((fileCoveredLines+fileCoveredBranches) / (fileLinesToCover +fileBranchesToCover) * 100).toFixed(2);
              if(overallCov+parseInt(options.variance) < minCov ) {
                uncoveredFiles[file.file] = {uncoveredLines:fileUncoveredLines, uncoveredBranches: fileUncoveredBranches}
              }
            }
          }
          
        })
        console.info(`Coverage check is completed.`);
        const overallCov = parseFloat((coveredLines+coveredBranches) / (linesToCover +branchesToCover) * 100).toFixed(2);
        console.info(`The overall coverage for changed and added lines is ${overallCov}`);
        generateHtmlReport(options.output,uncoveredFiles)
        if(overallCov+parseInt(options.variance) < minCov ) {
          exitCode = 1
          console.error(`The overall coverage (${overallCov}) is bellow the minimum of ${minCov}`);
        }
        if(exitCode > 0) {
          console.info(`Something went wrong, check the logs for more info.`);
        } else {
          console.info(`All the checks have passed succesfully.`);
        }
        process.exit(exitCode);
      } catch (error) {
        console.error(error);
        process.exit(1);
      }
      
    }).catch(function (err) {
        //In case the file can not be parsed, we assume there is nothing to cover
        console.info("There is nothing to analyse on this branch.");
        process.exit(0);
    });
  },
  validateOptions: function(options)
  {
    if(!options.project)
    {
      console.log("The project path is mandatory, no project path option passed on -p");
      process.exit(1);
    }
    
  }
}

module.exports = app;
